import sys
from argparse import ArgumentParser

from yamlargs.config import YAMLConfig


def create_parser(config: YAMLConfig) -> ArgumentParser:
    """
    Automatically creates a parser from all of the values specified in a config
    file. Will use the dot syntax for nested dictionaries.

    Parameters
    ----------
    config: YAMLConfig
        Config object

    Returns
    -------
    ArgumentParser
        Parser loaded up with all of the values specified in the config
    """
    key_pairs = config.keys()

    parser = ArgumentParser(
        description=f"""
This argument parser was autogenerated from the config file. This allows you to
overwrite specific YAML values on the fly. The options listed here do not
entail an exhaustive list of the things that you can configure. For more
information on possible kwargs, refer to the class definition of the object in
question. 
    """
    )

    parser.add_argument(f"config_file", help="YAML config file")
    for k in key_pairs:
        current = config.access(k)
        parser.add_argument(f"--{k}", type=type(current))

    return parser


def load_config_and_create_parser():
    """
    Loads a config and creates a parser.

    Use this to initialize everything when you want to add more arguments to
    the parser. Will load the config from the first sys.argv position.

    Returns
    -------
    YAMLConfig
        config object loaded from sys.argv[1]
    ArgumentParser
        Parser for the loaded config object
    """
    path = sys.argv[1]
    config = YAMLConfig.load(path)
    parser = create_parser(config)
    return config, parser


def parse_args_into_config(config, args, verbose=False):
    """"""
    for (k, v) in vars(args).items():
        if v and k != "config_file":
            config.set(k, v)

    if verbose:
        print(config)

    return config


def load_config_and_parse_cli(verbose=False):
    """
    Loads a config and parses all the command line arguments.

    One stop shop, will handle all of the intialization for you. Loads the
    config from sys.argv[1], creates an ArugmentParser, and then parses the
    specified args.

    Returns
    -------
    YAMLConfig
        Fully loaded config object
    """
    config, parser = load_config_and_create_parser()
    args = parser.parse_args()
    config = parse_args_into_config(config, args, verbose=verbose)
    return config



import sys
from typing import Any, Dict
from dataclasses import dataclass

import yaml


@dataclass(frozen=True)
class LazyConstructor:
    """
    An object that wraps a constructor in a lazy, configurable way.

    It turns the constructor into a dictionary like object that allows you to
    set and overwrite the default values like a dictionary, and then intialize
    kind of like a factory.

    LazyConstructor should be intialized with a class and a dictionary of key
    word arguemnts. It will then represent a lazy version of lambda *args:
    class(*args, **kwargs).

    If a LazyConstructor depends on a key word argument that is also a
    LazyConstructor, it will recursively intitialize it before intializing
    itself.

    >>> class Dice:
    ...       def __init__(self, value, max_value=6):
    ...               self.value = value
    ...               if self.value > max_value:
    ...                       print("Bad value")
    ...       def __str__(self):
    ...               return f"Dice({self.value})"
    ...       def __repr__(self):
    ...               return self.__str__()
    ...
    >>> c = LazyConstructor(Dice, {"max_value": 6})
    >>> c(1)
    Dice(1)
    >>> c(10)
    Bad value
    Dice(10)
    >>> c['max_value'] = 12
    >>> c(10)
    Dice(10)
    >>> print(c)
    {'class': <class '__main__.Dice'>, 'max_value': 12}
    >>> c = LazyConstructor(Dice, {"max_value": 6, "value": 1})
    >>> c()
    Dice(1)
    >>> c['value'] = 4
    >>> c()
    Dice(4)
    """

    _fn: Any
    _kwargs: Dict

    def __getitem__(self, key):
        if key == "class":
            return self._fn
        else:
            return self._kwargs.__getitem__(key)

    def __setitem__(self, key, value):
        if key == "class":
            self._fn = value
        else:
            self._kwargs.__setitem__(key, value)

    def __call__(self, *args, **kwargs):
        # If there are any LazyConstructors in the keyword arguments, intialize
        # them before intializing itself. Allows for a recursive chain of
        # intialization. Also copies over the specified kwargs.

        init_kwargs = {}
        for (k, v) in list(self._kwargs.items()) + list(kwargs.items()):
            if isinstance(v, LazyConstructor):
                init_kwargs[k] = v()
            else:
                init_kwargs[k] = v
        try:
            return self._fn(*args, **init_kwargs)
        except TypeError as e:
            raise TypeError(f"Error calling {self._fn.__name__}") from e

    def as_dict(self):
        kwargs = {}
        for (k, v) in self._kwargs.items():
            if isinstance(v, LazyConstructor):
                v = v.as_dict()
            else:
                v = repr(v)
            kwargs[k] = v
        d = {"lazyObject": "!" + self._fn.__name__ + "()", "kwargs": kwargs}
        return d

    def __repr__(self):
        d = self.as_dict()
        return str(d)

    def __str__(self):
        return self.__repr__()

    def keys(self):
        """Fake keys to act like a dictionary"""
        return ["class"] + list(self._kwargs.keys())

    def values(self):
        """Fake values to act like a dictionary"""
        return [self._fn] + list(self._kwargs.values())

    def items(self):
        """Fake items to act like a dictionary"""
        keys = self.keys()
        return [(k, self.__getitem__(k)) for k in keys]


@dataclass(frozen=True)
class LazyFunction(LazyConstructor):
    """
    An object that wraps a function in a lazy, configurable way.

    See LazyConstructor. Does the same thing, except it returns a function
    instead of an intialized value on call.

    >>> class Dice:
    ...       def __init__(self, value, max_value=6):
    ...               self.value = value
    ...               if self.value > max_value:
    ...                       print("Bad value")
    ...       def __str__(self):
    ...               return f"Dice({self.value})"
    ...       def __repr__(self):
    ...               return self.__str__()
    ...
    >>> c = LazyFunction(Dice, {"max_value": 6})
    >>> c()
    Dice
    >>> c()(10)
    Bad value
    Dice(10)
    """

    def as_dict(self):
        kwargs = {}
        for (k, v) in self._kwargs.items():
            if isinstance(v, LazyConstructor):
                v = v.as_dict()
            kwargs[k] = v
        d = {"lazyObject": "!" + self._fn.__name__, "kwargs": kwargs}
        return d

    def __call__(self, *args, **kwargs):
        base_fn = super().__call__

        # Allows for attributes to remain on wrapped functions, so if you wrap
        # a class, you can still access the classmethods
        attributes = filter(lambda x: x[0] != "_", dir(self._fn))
        attributes = {a: getattr(self._fn, a) for a in attributes}
        attributes["__call__"] = super().__call__

        return type(self._fn.__name__ + "(LazyFunction)", (object,), attributes)()


def _get_kwargs_from_node(loader, node):
    # TODO: This try except is a little weird. I feel like there should be a
    # way to check rather than the error
    try:
        kwargs = loader.construct_mapping(node)
    # No mapping specified, treat this an empty dict
    except yaml.constructor.ConstructorError:
        kwargs = {}
    return kwargs


def expose_module(mod):
    fns = dir(mod)
    non_private_fns = list(filter(lambda x: x[:2] != "__", fns))

    for fn in non_private_fns:
        obj_ = getattr(mod, fn)
        # TODO: Think of a way to still expose thigs without an explicit name
        # (lists)
        if hasattr(obj_, '__name__'):
            make_lazy(obj_)


def make_lazy(obj_, default_kwargs=None):
    make_lazy_function(obj_, default_kwargs=default_kwargs)
    make_lazy_constructor(obj_, default_kwargs=default_kwargs)


def make_lazy_constructor(type_, default_kwargs=None):
    """
    Exposes to YAML a lazy constructor for the object, so it can be referenced
    in a config. Exposes both the ability to create the object itself, and it's
    class as a type.

    >>> make_lazy_constructor(Dice)
    >>> data = yaml.load("dice: !Dice", yaml.UnsafeLoader)
    >>> data["dice"]()
    Dice()
    """

    name = type_.__name__

    def _constructor(loader, node):

        kwargs = _get_kwargs_from_node(loader, node)

        # Surely, there must be a better way to do this, right? I tried .update
        # but that was casuing weird issues.

        if default_kwargs is not None:
            for (k, v) in default_kwargs.items():
                if k not in kwargs.keys():
                    kwargs[k] = v

        return LazyConstructor(type_, kwargs)

    def _type_constructor(loader, node):
        return lambda: type_

    yaml.add_constructor("!" + name + "()", _constructor)


def make_lazy_function(type_, default_kwargs=None):
    """
    Exposes to YAML a lazy function, so it can be referenced
    in a config.

    >>> def fn(x, b=1):
    ...     return x**2 + b
    >>> make_lazy_function(fn)
    >>> data = yaml.load("my_fn: !fn\n b: 2", yaml.UnsafeLoader)
    >>> myfn = data["my_fn"]()
    >>> myfn(2)
    6
    """

    name = type_.__name__

    def _constructor(loader, node):
        kwargs = _get_kwargs_from_node(loader, node)

        if default_kwargs is not None:
            for (k, v) in default_kwargs.items():
                if k not in kwargs.keys():
                    kwargs[k] = v

        return LazyFunction(type_, kwargs)

    yaml.add_constructor("!" + name, _constructor)



import yaml
from yamlargs.config import YAMLConfig

from yamlargs import make_lazy_constructor, make_lazy_function


class Dice:
    def __init__(self, value, max_value=6):
        self.value = value
        self.max_value = max_value
        if self.value > self.max_value:
            print("Bad value")

    def __str__(self):
        return f"Dice({self.value})"

    def __repr__(self):
        return self.__str__()


class DiceCup:
    def __init__(self, dice=Dice(1, max_value=10)):
        self.dice = dice


def fake_fn(x, n=10):
    return x + n


make_lazy_constructor(Dice)
make_lazy_function(Dice)

make_lazy_constructor(DiceCup)

make_lazy_function(fake_fn)


def test_load():
    config = YAMLConfig.load("test/example.yml")
    assert config["load"]["dice"]["value"] == 123
    assert config["fn"]["n"] == 100


def test_dot_access():
    config = YAMLConfig.load("test/example.yml")
    assert config.access("load.dice.value") == 123
    assert config.access("fn.n") == 100


def test_dot_set():
    config = YAMLConfig.load("test/example.yml")
    config.set("load.dice.value", 543)
    config.set("fn.n", 543)
    assert config.access("load.dice.value") == 543
    assert config.access("fn.n") == 543


def test_keys():
    config = YAMLConfig.load("test/example.yml")
    assert config.keys() == [
        "init.class",
        "init.value",
        "load.class",
        "load.dice.class",
        "load.dice.value",
        "fn.class",
        "fn.n",
    ]


def test_json():
    config = YAMLConfig.load("test/example.yml")
    data = config.to_json()
    assert data["init"] == {"lazyObject": "!Dice", "kwargs": {"value": 49}}
    assert config == YAMLConfig.from_json(config.to_json())


def test_yaml():
    config = YAMLConfig.load("test/example.yml")
    assert config == YAMLConfig.from_yaml(config.to_yaml())



from yamlargs.lazy import (
    make_lazy_constructor,
    make_lazy_function,
    make_lazy,
    expose_module,
)
from yamlargs.config import YAMLConfig
from yamlargs.parser import load_config_and_parse_cli, load_config_and_create_parser



import yaml

from yamlargs.lazy import LazyConstructor, make_lazy_constructor, make_lazy_function


class Dice:
    def __init__(self, value, max_value=6):
        self.value = value
        self.max_value = max_value
        if self.value > self.max_value:
            print("Bad value")

    def __str__(self):
        return f"Dice({self.value})"

    def __repr__(self):
        return self.__str__()


class DiceCup:
    def __init__(self, dice=Dice(1, max_value=10)):
        self.dice = dice


def test_kwargs():
    c = LazyConstructor(Dice, {"max_value": 10})
    assert c(0).max_value == 10

    c = LazyConstructor(Dice, {"max_value": -1})
    assert c(0).max_value == -1


def test_kwargs_overwrite():
    c = LazyConstructor(Dice, {"max_value": 30})
    c["max_value"] = -1

    assert c(1).max_value == -1


def test_args():
    c = LazyConstructor(Dice, {"value": 1})
    assert c().value == 1
    assert c(max_value=4).value == 1
    assert c(max_value=4).max_value == 4


def test_args_overwrite():
    c = LazyConstructor(Dice, {"value": 1})
    c["value"] = 5
    assert c().value == 5


def test_lazy_as_kwarg():
    lazy_dice = LazyConstructor(Dice, {"value": 123, "max_value": 1234})
    lazy_cup = LazyConstructor(DiceCup, {"dice": lazy_dice})

    cup = lazy_cup()
    assert cup.dice.value == 123
    assert cup.dice.max_value == 1234


def test_yaml_loading_no_args():
    make_lazy_constructor(Dice)
    data = yaml.load("load: !Dice()", yaml.UnsafeLoader)
    assert data["load"](1).value == 1


def test_yaml_loading():
    make_lazy_constructor(Dice)
    data = yaml.load("load: !Dice()\n value: 1", yaml.UnsafeLoader)
    assert data["load"]().value == 1


def test_yaml_loading_with_kwargs():
    make_lazy_constructor(Dice, {"max_value": 10})
    data = yaml.load("load: !Dice()\n value: 1", yaml.UnsafeLoader)
    assert data["load"]().value == 1
    assert data["load"]().max_value == 10


def test_yaml_loading_lazy_recursive():
    make_lazy_constructor(Dice)
    make_lazy_constructor(DiceCup)
    data = yaml.load(
        "load: !DiceCup()\n dice: !Dice()\n  value: 123", yaml.UnsafeLoader
    )
    assert data["load"]().dice.value == 123


def test_yaml_function_with_kwargs():
    def fn(x, b=1):
        return x**2 + b

    make_lazy_function(fn)
    data = yaml.load("my_fn: !fn\n b: 2", yaml.UnsafeLoader)
    myfn = data["my_fn"]()
    assert myfn(2) == 6


def test_yaml_function():
    def fn(x, b=1):
        return x**2 + b

    make_lazy_function(fn)
    data = yaml.load("my_fn: !fn", yaml.UnsafeLoader)
    myfn = data["my_fn"]()
    assert myfn(2) == 5


def test_yaml_function_recurse():
    def fn(x, b=1):
        return x**2 + b

    make_lazy_function(fn)
    data = yaml.load("cup: !DiceCup()\n dice: !fn\n  b: 2", yaml.UnsafeLoader)
    cup = data["cup"]()
    assert cup.dice(2) == 6



from dataclasses import dataclass
from argparse import ArgumentParser
from typing import Dict, Any

import yaml

from yamlargs.lazy import LazyConstructor


@dataclass(frozen=True)
class YAMLConfig:
    """
    Config object.
    """

    data: Dict

    @classmethod
    def load(cls, path: str) -> "YAMLConfig":
        """
        Loads the config into a YAMLConfig object.

        This basically only exists to remind you to use yaml.UnsafeLoader.

        Parameters
        ----------
        path: str
            path to yaml config file

        Returns
        -------
        config: YAMLConfig
            Initialized config instance
        """
        data = yaml.load(open(path, "r"), yaml.UnsafeLoader)
        return cls(data)

    @classmethod
    def from_json(cls, data):
        yaml_data = _dict_to_yaml(data)
        data = yaml.load(yaml_data, yaml.UnsafeLoader)
        return cls(data)

    @classmethod
    def from_yaml(cls, yaml_str):
        data = yaml.load(yaml_str, yaml.UnsafeLoader)
        return cls(data)

    def access(self, access_str: str):
        return _dot_access(self.data, access_str)

    def set(self, access_str: str, new_value: Any):
        _dot_set(self.data, access_str, new_value)

    def keys(self):
        return _get_all_keys(self.data)

    def __getitem__(self, idx):
        return self.data[idx]

    def to_json(self):
        """
        Converts the config into JSON data for logging purposes
        """
        new_data = {}
        for (k, v) in self.data.items():
            if isinstance(v, LazyConstructor):
                new_data[k] = v.as_dict()
            else:
                new_data[k] = v
        return new_data

    def to_yaml(self):
        """
        Converts the config into JSON data for logging purposes
        """
        yaml_data = ""
        json_data = self.to_json()
        return _dict_to_yaml(json_data)


def _dict_to_yaml(data, indent=""):
    """
    Helper function for recursively turning the json data into yaml strings.
    """
    yaml_data = ""
    for (k, v) in data.items():
        if isinstance(v, dict) and "lazyObject" in v.keys():
            name = v["lazyObject"]
            yaml_data += f"{indent}{k}: {name}\n"
            yaml_data += _dict_to_yaml(v["kwargs"], indent + " " * 4)
        else:
            yaml_data += f"{indent}{k}: {v}\n"
    return yaml_data


def _get_all_keys(config):
    keys = []
    for (k, v) in config.items():
        # isinstance(LazyConstructor, "keys") actually returns true, so we need
        # to double check that it is specifically not a LazyConstructor
        if hasattr(v, "keys") and v != LazyConstructor:
            subkeys = _get_all_keys(v)
            for sk in subkeys:
                keys.append(".".join([k, sk]))
        else:
            keys.append(k)

    return keys


def _dot_access(nested_dict, dot_key):
    keys = dot_key.split(".")
    d = nested_dict
    for k in keys:
        d = d[k]
    return d


def _dot_set(nested_dict, dot_key, value):
    keys = dot_key.split(".")
    d = nested_dict
    for k in keys[:-1]:
        d = d[k]
    d[keys[-1]] = value
    return d
